@@ -1,6 +1,6 @@
 ;;; llm-claude.el --- llm module for integrating with Claude -*- lexical-binding: t; package-lint-main-file: "llm.el"; -*-
 
-;; Copyright (c) 2024  Free Software Foundation, Inc.
+;; Copyright (c) 2024-2025  Free Software Foundation, Inc.
 
 ;; Author: Andrew Hyatt <ahyatt@gmail.com>
 ;; Homepage: https://github.com/ahyatt/llm
@@ -43,64 +43,86 @@
   (unless (llm-claude-key provider)
     (error "No API key provided for Claude")))
 
-(defun llm-claude--tool-call (call)
-  "A Claude version of a function spec for CALL."
-  `(("name" . ,(llm-function-call-name call))
-    ("description" . ,(llm-function-call-description call))
-    ("input_schema" . ,(llm-provider-utils-openai-arguments
-                        (llm-function-call-args call)))))
+(defun llm-claude--tool-call (tool)
+  "A Claude version of a function spec for TOOL."
+  `(:name ,(llm-tool-function-name tool)
+          :description ,(llm-tool-function-description tool)
+          :input_schema ,(llm-provider-utils-openai-arguments
+                          (llm-tool-function-args tool))))
 
 (cl-defmethod llm-provider-chat-request ((provider llm-claude) prompt stream)
-  (let ((request `(("model" . ,(llm-claude-chat-model provider))
-                   ("stream" . ,(if stream t :json-false))
+  (let ((request
+          `(:model ,(llm-claude-chat-model provider)
+                   :stream ,(if stream t :false)
                    ;; Claude requires max_tokens
-                   ("max_tokens" . ,(or (llm-chat-prompt-max-tokens prompt) 4096))
-                   ("messages" .
-                    ,(mapcar (lambda (interaction)
-                               `(("role" . ,(pcase (llm-chat-prompt-interaction-role interaction)
-                                              ('function 'user)
-                                              ('assistant 'assistant)
-                                              ('user 'user)))
-                                 ("content" .
-                                  ,(if (llm-chat-prompt-interaction-function-call-results interaction)
-                                       (mapcar (lambda (result)
-                                                 `(("type" . "tool_result")
-                                                   ("tool_use_id" .
-                                                    ,(llm-chat-prompt-function-call-result-call-id
-                                                      result))
-                                                   ("content" .
-                                                    ,(llm-chat-prompt-function-call-result-result result))))
-                                               (llm-chat-prompt-interaction-function-call-results interaction))
-                                     (llm-chat-prompt-interaction-content interaction)))))
+                   :max_tokens ,(or (llm-chat-prompt-max-tokens prompt) 4096)
+                   :messages
+                   ,(vconcat
+                     (mapcar (lambda (interaction)
+                               `(:role  ,(pcase (llm-chat-prompt-interaction-role interaction)
+                                           ('tool_results "user")
+                                           ('tool_use "assistant")
+                                           ('assistant "assistant")
+                                           ('user "user"))
+                                        :content
+                                        ,(cond ((llm-chat-prompt-interaction-tool-results interaction)
+                                                (vconcat (mapcar (lambda (result)
+                                                                   `(:type "tool_result"
+                                                                           :tool_use_id
+                                                                           ,(llm-chat-prompt-tool-result-call-id result)
+                                                                           :content
+                                                                           ,(llm-chat-prompt-tool-result-result result)))
+                                                                 (llm-chat-prompt-interaction-tool-results interaction))))
+                                               ((llm-multipart-p (llm-chat-prompt-interaction-content interaction))
+                                                (llm-claude--multipart-content
+                                                 (llm-chat-prompt-interaction-content interaction)))
+                                               (t
+                                                (llm-chat-prompt-interaction-content interaction)))))
                              (llm-chat-prompt-interactions prompt)))))
         (system (llm-provider-utils-get-system-prompt prompt)))
-    (when (llm-chat-prompt-functions prompt)
-      (push `("tools" . ,(mapcar (lambda (f) (llm-claude--tool-call f))
-                                 (llm-chat-prompt-functions prompt))) request))
+    (when (llm-chat-prompt-tools prompt)
+      (plist-put request :tools
+                 (vconcat (mapcar (lambda (f) (llm-claude--tool-call f))
+                                  (llm-chat-prompt-tools prompt)))))
     (when (> (length system) 0)
-      (push `("system" . ,system) request))
+      (plist-put request :system system))
     (when (llm-chat-prompt-temperature prompt)
-      (push `("temperature" . ,(llm-chat-prompt-temperature prompt)) request))
-    (append request (llm-chat-prompt-non-standard-params prompt))))
+      (plist-put request :temperature (llm-chat-prompt-temperature prompt)))
+    (append request (llm-provider-utils-non-standard-params-plist prompt))))
 
-(cl-defmethod llm-provider-extract-function-calls ((_ llm-claude) response)
+(defun llm-claude--multipart-content (content)
+  "Return CONTENT as a list of Claude multipart content."
+  (vconcat (mapcar (lambda (part)
+                     (cond ((stringp part)
+                            `(:type "text"
+                                    :text ,part))
+                           ((llm-media-p part)
+                            `(:type "image"
+                                    :source (:type "base64"
+                                                   :media_type ,(llm-media-mime-type part)
+                                                   :data ,(base64-encode-string (llm-media-data part) t))))
+                           (t
+                            (error "Unsupported multipart content: %s" part))))
+                   (llm-multipart-parts content))))
+
+(cl-defmethod llm-provider-extract-tool-uses ((_ llm-claude) response)
   (let ((content (append (assoc-default 'content response) nil)))
     (cl-loop for item in content
              when (equal "tool_use" (assoc-default 'type item))
-             collect (make-llm-provider-utils-function-call
+             collect (make-llm-provider-utils-tool-use
                       :id (assoc-default 'id item)
                       :name (assoc-default 'name item)
                       :args (assoc-default 'input item)))))
 
-(cl-defmethod llm-provider-populate-function-calls ((_ llm-claude) prompt calls)
+(cl-defmethod llm-provider-populate-tool-uses ((_ llm-claude) prompt tool-uses)
   (llm-provider-utils-append-to-prompt
    prompt
-   (mapcar (lambda (call)
-             `((type . "tool_use")
-               (id . ,(llm-provider-utils-function-call-id call))
-               (name . ,(llm-provider-utils-function-call-name call))
-               (input . ,(llm-provider-utils-function-call-args call))))
-           calls)))
+   (vconcat (mapcar (lambda (call)
+                      `(:type "tool_use"
+                              :id ,(llm-provider-utils-tool-use-id call)
+                              :name ,(llm-provider-utils-tool-use-name call)
+                              :input ,(llm-provider-utils-tool-use-args call)))
+                    tool-uses))))
 
 (cl-defmethod llm-provider-chat-extract-result ((_ llm-claude) response)
   (let ((content (aref (assoc-default 'content response) 0)))
@@ -129,6 +151,9 @@
                          (when (equal type "text_delta")
                            (funcall msg-receiver (assoc-default 'text delta))))))))))
 
+(cl-defmethod llm-provider-collect-streaming-tool-uses ((_ llm-claude) data)
+  (llm-provider-utils-openai-collect-streaming-tool-uses data))
+
 (cl-defmethod llm-provider-headers ((provider llm-claude))
   `(("x-api-key" . ,(if (functionp (llm-claude-key provider))
                         (funcall (llm-claude-key provider))
@@ -153,16 +178,16 @@
   "Claude")
 
 (cl-defmethod llm-capabilities ((_ llm-claude))
-  (list 'streaming 'function-calls))
+  (list 'streaming 'function-calls 'image-input))
 
 (cl-defmethod llm-provider-append-to-prompt ((_ llm-claude) prompt result
-                                             &optional func-results)
+                                             &optional tool-use-results)
   ;; Claude doesn't have a 'function role, so we just always use assistant here.
   ;; But if it's a function result, it considers that a 'user response, which
   ;; needs to be sent back.
-  (llm-provider-utils-append-to-prompt prompt result func-results (if func-results
-                                                                      'user
-                                                                    'assistant)))
+  (llm-provider-utils-append-to-prompt prompt result tool-use-results (if tool-use-results
+                                                                          'user
+                                                                        'assistant)))
 
 
 (provide 'llm-claude)
